from dataclasses import dataclass
from datetime import datetime

import sqlite3
import sys


MOVIEFILE = "movies.txt"
DBFILE = "db/im.db"
TBLTEMPLATEFILE = "wiki_tbl_template.txt"
LSTTEMPLATEFILE = "wiki_lst_template.txt"
WIKITBLOUTPUT = "wiki_tbl_{}.txt".format(datetime.now().strftime("%Y%m%d%H%M%S"))
WIKILSTOUTPUT = "wiki_lst_{}.txt".format(datetime.now().strftime("%Y%m%d%H%M%S"))
MOVIESOUTPUT = "movies_{}.txt".format(datetime.now().strftime("%Y%m%d%H%M%S"))


# imdb_id, media_type, original_media_type, primary_title, original_title, is_adult, start_year, end_year, runtime, genres
# imdb_id, media_type, title, earliest_year, latest_year, wiki_entry
INSERT = "insert into movies values (?, ?, ?, ?, ?, ?)"
UPDATE = "update movies set imdb_id = ?, media_type = ?, title = ?, earliest_year = ?, latest_year = ?, wiki_entry = ? where title = ?"
SELECT_MATCHES = 'select distinct i.imdb_id, m.media_type, m.title, i.primary_title, i.original_title, i.start_year, i.end_year, i.runtime, i.genres from movies m join imdb_records i on (m.title = i.original_title or m.title = i.primary_title) and i.media_type like "%"||m.media_type||"%" and (m.imdb_id is null or m.imdb_id = i.imdb_id) and (i.start_year is null or (m.earliest_year is null and m.latest_year is null) or (m.earliest_year is null and i.start_year <= m.latest_year) or (m.latest_year is null and i.start_year >= m.earliest_year) or (i.start_year between m.earliest_year and m.latest_year)) where m.wiki_entry is null order by m.title collate nocase'
SELECT_MISSING = "select * from movies where wiki_entry is null"


inserts = []
updates = []
wiki_tbl_entries = []
wiki_lst_entries = []


@dataclass
class Record():
    imdb_id : str
    media_type : str
    user_title : str
    primary_title : str
    original_title : str
    start_year : int
    end_year : int
    runtime : int
    genres : list
    wiki_entry : str = None

    def __post_init__(self):
        self.wiki_entry = "<tr><td><a title='{} auf imdb.com' href='https://www.imdb.com/title/{}/' target='_blank' rel='noopener'>{}</a></td><td>{}</td><td>{}</td><td>{}</td></tr>"\
                          .format("" if self.user_title is None else self.user_title, 
                                  "" if self.imdb_id is None else self.imdb_id, 
                                  "" if self.user_title is None else self.user_title, 
                                  "" if self.start_year is None else self.start_year, 
                                  "" if self.runtime is None else self.runtime, 
                                  "" if self.genres is None else ", ".join(self.genres))

    def get_insert_values(self) -> tuple:
        # imdb_id, media_type, title, earliest_year, latest_year, wiki_entry
        return (self.imdb_id, self.media_type, self.user_title, 
                self.start_year, self.end_year, 
                None if self.imdb_id is None else self.wiki_entry)

    def get_update_values(self) -> tuple:
        # imdb_id, media_type, title, earliest_year, latest_year, wiki_entry
        return (self.imdb_id, self.media_type, self.user_title, 
                self.start_year, self.end_year, 
                None if self.imdb_id is None else self.wiki_entry, 
                # plus update where clause param
                self.user_title)


def bulk_import() -> None:
    try:
        conn = sqlite3.connect(DBFILE)
    except:
        sys.exit()

    with conn:
        curs = conn.cursor()
        curs.executemany(INSERT, inserts)
        conn.commit()
        curs.close()


def bulk_update() -> None:
    try:
        conn = sqlite3.connect(DBFILE)
    except:
        sys.exit()

    with conn:
        curs = conn.cursor()
        curs.executemany(UPDATE, updates)
        conn.commit()
        curs.close()


def read_file() -> str:
    with open(MOVIEFILE) as movie_file:
        lines = movie_file.readlines()
    return lines


def create_records(lines: list) -> None:
    for line in lines:
        line = line.strip("\n")
        line = line.strip(" ")
        fields = line.split(";")
        while len(fields) < 6: fields.append(None)
        if len(fields) > 6:
            # re-import fields:
            # i.imdb_id, m.media_type, m.title, i.primary_title, i.original_title, 
            # i.start_year, i.end_year, i.runtime, i.genres
            # matching on columns:
            # imdb_id, media_type, title, earliest_year, latest_year, wiki_entry
            movie = Record(imdb_id = fields[0], media_type = fields[1], 
                           user_title = fields[2], 
                           primary_title = fields[3], original_title = fields[4], 
                           start_year = fields[5], end_year = fields[6], 
                           runtime = fields[7], genres = fields[8].split(","))
            updates.append(movie.get_update_values())
            wiki_tbl_entries.append(movie.wiki_entry)
        else:
            # plain user input file:
            # Alma;Serie;2020;2023
            # matching on columns:
            # imdb_id, media_type, title, earliest_year, latest_year, wiki_entry
            movie = Record(imdb_id = None, media_type = fields[1], 
                           user_title = fields[0], 
                           primary_title = fields[0], original_title = fields[0], 
                           start_year = fields[2], end_year = fields[3], 
                           runtime = None, genres = None)
            inserts.append(movie.get_insert_values())


def create_movies(results: list) -> list:
    movies = []
    for rec in results:
        movie = Record(imdb_id = rec[0], media_type = rec[1], 
                       user_title = rec[2], 
                       primary_title = rec[3], original_title = rec[4], 
                       start_year = rec[5], end_year = rec[6], 
                       runtime = rec[7], genres = rec[8].split(","))
        movies.append(movie)
    return movies


def get_matches() -> list:
    try:
        conn = sqlite3.connect(DBFILE)
    except:
        sys.exit()

    with conn:
        curs = conn.cursor()
        results = curs.execute(SELECT_MATCHES)
        movies = create_movies(results)
        curs.close()

    return movies


def create_missing_list(results: list) -> None:
    for rec in results:
        # imdb_id, media_type, title, earliest_year, latest_year, wiki_entry
        wiki_lst_entries.append(Record(imdb_id = None, media_type = rec[1], 
                                       user_title = rec[2], 
                                       primary_title = None, original_title = None, 
                                       start_year = rec[3], end_year = rec[4], 
                                       runtime = None, genres = None))


def load_missing() -> None:
    try:
        conn = sqlite3.connect(DBFILE)
    except:
        sys.exit()

    with conn:
        curs = conn.cursor()
        results = curs.execute(SELECT_MISSING)
        movies = create_missing_list(results)
        curs.close()


def get_template(template_name: str) -> str:
    with open(template_name) as template_file:
        lines = template_file.readlines()
    return "\n".join(lines)


def write_wiki_tbl_entries(template: str) -> None:
    with open(WIKITBLOUTPUT, 'w') as wiki_tbl_entries_file:
        wiki_tbl_entries_file.write(template.replace("TABLE_CONTENT", 
                                                     "\n".join(wiki_entry for wiki_entry in wiki_tbl_entries)
                                                    )
                                   )


def write_wiki_lst_entries(template: str) -> None:
    with open(WIKILSTOUTPUT, 'w') as wiki_lst_entries_file:
        wiki_lst_entries_file.write(template.replace("LIST_CONTENT", 
                                                     "\n".join("<li class='null'>{}, {}, from {}, to {}</li>"
                                                               .format(m.user_title, m.media_type, 
                                                                       str(m.start_year), str(m.end_year))
                                                               for m in wiki_lst_entries)
                                                    )
                                   )


def write_movies_file(movies: list) -> None:
    # re-import fields:
    # i.imdb_id, m.media_type, m.title, i.primary_title, i.original_title, 
    # i.start_year, i.end_year, i.runtime, i.genres
    with open(MOVIESOUTPUT, 'w') as movies_file:
        movies_file.write("\n".join( [";".join( 
                                                [m.imdb_id, m.media_type, m.user_title, 
                                                 m.primary_title, m.original_title, 
                                                 str(m.start_year), str(m.end_year), 
                                                 str(m.runtime), 
                                                 ",".join(m.genres)]
                                              ) for m in movies
                                     ]
                                   )
                         )


# store input data
lines = read_file()
create_records(lines)
bulk_import()
inserts.clear()

bulk_update()
updates.clear()

# get potential matches
movies = get_matches()
write_movies_file(movies)

print("matches for imported movies:")
for movie in movies:
    print(movie)

# write table output
template = get_template(TBLTEMPLATEFILE)
write_wiki_tbl_entries(template)
wiki_tbl_entries.clear()

# write list output
load_missing()
template = get_template(LSTTEMPLATEFILE)
write_wiki_lst_entries(template)
wiki_lst_entries.clear()


